import numpy as np
import h5py
import logging
import time

logger = logging.getLogger(__name__)


def get_position(index, shape, scan_rank):
    """
    Returns the position in the scan associated with the given index

            Parameters:
                    index (int): Flattened index of scan point
                    shape (array): Shape of dataset of interest
                    scan_rank (int): Rank of scan (must be <= len(shape))

            Returns:
                    position (tuple): Ndimensional position in scan associated with the index
    Examples
    --------

    >>> utils.get_position(0, [3,4,5], 2)
    (0,0)

    """

    scan_shape = shape[:scan_rank]
    pos = np.unravel_index(index, scan_shape)

    return pos


def get_position_snake(index, shape, scan_rank):
    """
    Returns the position in a snake scan associated with the given index

            Parameters:
                    index (int): Flattened index of scan point
                    shape (array): Shape of dataset of interest
                    scan_rank (int): Rank of scan (must be <= len(shape))

            Returns:
                    position (tuple): Ndimensional position in scan associated with the index
    Examples
    --------

    >>> utils.get_position_snake(0, [3,4,5], 2)
    (0,0)

    """
    non_snake = get_position(index, shape, scan_rank)
    out = [*non_snake]
    sum = non_snake[0]
    for i in range(1, len(out)):
        if sum % 2 == 1:
            out[i] = shape[i] - non_snake[i] - 1

        sum += out[i]

    return tuple(out)


def get_row_slice(index, shape, scan_rank):
    """
    Returns a slice tuple corresponding to the row of the shape that contains the index

            Parameters:
                    index (int): Flattened index of scan point
                    shape (array): Shape of dataset of interest
                    scan_rank (int): Rank of scan (must be <= len(shape))

            Returns:
                    row_slice (tuple): tuple of slices (same length as scan_rank) corresponding to the row containing the index
    Examples
    --------


    >>> utils.get_row_slice(3, [3,4,5], 2)
    (slice(0,1,None),slice(None,None,None)

    """
    pos = get_position(index, shape, scan_rank)

    rank = len(shape)
    slices = [slice(0, None, 1)] * rank

    for i in range(len(pos)):
        slices[i] = slice(pos[i], pos[i] + 1)

    slices[scan_rank - 1] = slice(None)

    return tuple(slices[:scan_rank])


def create_dataset(data, scan_maxshape, fh, path, **kwargs):
    """
    Convenience method to create a hdf5 dataset corresponding to data being the first dataset in a scan with shape scan_maxshape

            Parameters:
                    data (numpy array): First dataset to save in hdf5, corresponding to the first point in the scan
                    scan_maxshape (array): Shape of the scan, for example [1000] for a stack or [100,100] for a grid
                    fh (h5py File or Group): File or Group to create dataset in
                    path (str): path to save the dataset under
                    kwargs: forwared to the h5py create dataset method allowing chunking and compression to be specified

            Returns:
                    dataset (h5py Dataset): the result dataset

    """

    maxshape = scan_maxshape + data.shape
    shape = [1] * len(scan_maxshape) + list(data.shape)
    r = data.reshape(shape)

    # if chunks not set use data shape
    if "chunks" not in kwargs:

        if data.size < 10:
            c = [1 if i is None else i for i in maxshape]
            kwargs["chunks"] = tuple(c)
        else:
            kwargs["chunks"] = tuple(shape)

    return fh.create_dataset(path, data=r, maxshape=maxshape, **kwargs)


def append_data(data, slice_metadata, dataset):
    """
    Convenience method to append data to a hdf5 dataset generated in create_dataset

            Parameters:
                    data (numpy array): dataset to save in hdf5, corresponding to the n-th point in the scan
                    slice_metadata (array): Slice describing current position in scan
                    dataset (h5py Dataset): Dataset to set slice in

    """
    ds = tuple(slice(0, s, 1) for s in data.shape)
    fullslice = slice_metadata + ds
    current = dataset.shape
    new_shape = tuple(max(s.stop, c) for (s, c) in zip(fullslice, current))
    if np.any(new_shape > current):
        dataset.resize(new_shape)
    dataset[fullslice] = data


def copy_nexus_axes(nxd_in, nxd_out, scan_rank, frame_axes=None):
    """
    Copy the axes and associated attributes from the input NXdata to the output NXdata

            Parameters:
                    nxd_in (h5py Group): input NXdata to copy from
                    nxd_out (h5py Group): output NXdata to copy to
                    scan_rank (int): number of scan dimensions
                    frame_axes(String array): additional axes for data frame

    """
    axout = ["."] * scan_rank
    if frame_axes is not None:
        axout = axout + frame_axes

    if "axes" in nxd_in.attrs:
        axin = nxd_in.attrs["axes"]
        for i in range(scan_rank):
            a = axin[i]
            ax = a.decode("utf-8") if hasattr(a, "decode") else a
            axout[i] = ax
            if ax in nxd_in:
                nxd_out.copy(nxd_in[ax], nxd_out, ax)
                at = ax + "_indices"
                if at in nxd_in.attrs:
                    nxd_out.attrs[at] = nxd_in.attrs[at]

    nxd_out.attrs["axes"] = axout


def create_nxdata(e, name, default=True):
    """
    Create NXdata group
            Parameters:
                    e (h5py Group): group to create NXdata in
                    name (str): Name of NXdata group
                    default (boolean): add default tag to e pointing at new NXdata

    """
    return _create_nexus_with_default(e, name, "NXdata", default=default)


def create_nxentry(fh, name, default=True):
    """
    Create NXentry group
            Parameters:
                    fh (h5py Group): group to create NXentry in
                    name (str): Name of NXentry group
                    default (boolean): add default tag to parent pointing at new NXentry

    """
    return _create_nexus_with_default(fh, name, "NXentry", default=default)


def _create_nexus_with_default(g, name, nxclass, default=True):
    d = g.create_group(name)
    d.attrs["NX_class"] = nxclass
    if default:
        g.attrs["default"] = name

    return d


def check_file_readable(path, datasets, timeout=10, retrys=5):
    """
    Check all datasets in a file can be read, used to determine if all files are readable if main file contains links

            Parameters:
                    path (str): path to hdf5 file being checked
                    datasets (array): List of paths to datasets in the file

    """
    start = time.time()
    dif = time.time() - start

    while dif < timeout:
        try:
            with h5py.File(path, "r", libver="latest", swmr=True) as fh:
                for d in datasets:
                    fh[d]
                return True

        except Exception as e:
            logger.debug("Reading failed, retrying " + str(e))
            time.sleep(timeout / retrys)
            dif = time.time() - start

    logger.error("Could not read file " + path)
    return False


def convert_stack_to_grid(slices, scan_shape, snake=False):

    index = slices[0].start

    if not snake:
        pos = get_position(index, scan_shape, len(scan_shape))
    else:
        pos = get_position_snake(index, scan_shape, len(scan_shape))

    return _get_slices_from_position(pos, slices)


def _get_slices_from_position(pos, slices):
    slices_out = [slice(p, p + 1, 1) for p in pos]
    slices_out += slices[1:]
    return slices_out



import numpy as np
import time

import logging

logger = logging.getLogger(__name__)


class KeyFollower:
    """Iterator for following key datasets in hdf5 files

    Parameters
    ----------
    h5file: h5py.File
        Instance of h5py.File object. Choose the file containing data you wish
        to follow.

    keypaths: list
        A list of paths (as strings) to key datasets in the hdf5 file. Can also be
        the path to groups, but the groups must contain only key datasets

    timeout: int (optional)
        The maximum time allowed for a dataset to update before the timeout
        termination condition is trigerred and iteration is halted. If a value
        is not set this will default to 10 seconds.

    finished_dataset: string (optional)
        Path to a scalar hdf5 dataset which is zero when the file is being
        written to and non-zero when the file is complete. Used to stop
        the iterator without waiting for the timeout



    Examples
    --------


    >>> # open hdf5 file using context manager with swmr mode activated
    >>> with h5py.File("/home/documents/work/data/example.h5", "r", swmr = True) as f:
    >>>     # create an instance of the Follower object to iterate through
    >>>     kf = KeyFollower(f,
    >>>                   ['/path/to/key/one', '/path/to/key/two'],
    >>>                   timeout = 10,
    >>>                   finished_dataset = "/path/to/finished")
    >>>     # iterate through the iterator as with a standard iterator/generator object
    >>>     for key in kf:
    >>>         print(key)



    """

    def __init__(self, h5file, keypaths, timeout=10, finished_dataset=None):
        self.h5file = h5file
        self.current_key = -1
        self.current_max = -1
        self.timeout = timeout
        self.start_time = None
        self.key_datasets = keypaths
        self.finished_dataset = finished_dataset
        self._finish_tag = False
        self._check_successful = False
        self.scan_rank = -1
        self._prelim_finished_check = False
        self.maxshape = None

    def __iter__(self):
        return self

    def check_datasets(self):
        if self._check_successful:
            return

        rank = -1

        key_list = self._get_key_list()

        for k in key_list:
            # do some exception checking here
            tmp = self.h5file[k]
            r = self._get_rank(tmp.maxshape)

            if rank == -1:
                rank = r

            if rank != -1 and rank != r:
                raise RuntimeError("Key datasets must have the same rank!")

            if self.maxshape is None:
                self.maxshape = tmp.maxshape[:rank]
            else:
                if np.all(self.maxshape != tmp.maxshape[:rank]):
                    logger.warning("Max shape not consistent in keys")

        if self.finished_dataset is not None:
            # just check read here
            tmp = self.h5file[self.finished_dataset]

        self.scan_rank = rank
        logger.debug("Dataset checks passed")

    def _get_key_list(self):
        key_list = self.key_datasets
        if len(key_list) == 1 and not hasattr(self.h5file[key_list[0]], "shape"):
            k0 = key_list[0]
            ks = []
            for k in self.h5file[k0]:
                ks.append(k0 + "/" + k)
            return ks
        else:
            return key_list

    def _get_rank(self, max_shape):
        rank = len(max_shape)
        rank_cor = 0
        for i in range(len(max_shape) - 1, -1, -1):
            if max_shape[i] == 1:
                rank_cor = rank_cor + 1
            else:
                break

        return rank - rank_cor

    def __next__(self):

        if self.current_key < self.current_max:
            self.current_key += 1
            return self.current_key

        self._timer_reset()
        while not self._is_next():
            time.sleep(self.timeout / 20.0)
            if self.is_finished():
                self._finish_tag = True
                raise StopIteration

        self.current_key += 1
        return self.current_key

    def reset(self):
        """Reset the iterator to start again from index 0"""
        self.current_key = -1
        self._finish_tag = False
        self._prelim_finished_check = False

    def _timer_reset(self):
        # Hidden method, restarts timer for timeout method
        self.start_time = time.time()

    def _is_next(self):

        karray = self._get_keys()
        if not karray:
            return False

        if len(karray) == 1:
            merged = karray[0]
        else:
            max_size = max([x.size for x in karray])

            merged = np.zeros(max_size)
            first = karray[0]
            merged[: first.size] = merged[: first.size] + first
            for k in karray[1:]:
                padded = np.zeros(max_size)
                padded[: k.size] = k
                merged = merged * padded

        new_max = np.argmax(merged == 0) - 1

        if new_max < 0 and merged[0] != 0:
            # all keys non zero
            new_max = merged.size - 1

        if self.current_max == new_max:
            return False

        self.current_max = new_max
        return True

    def _get_keys(self):
        kds = []
        for key_path in self._get_key_list():
            dataset = self.h5file[key_path]
            dataset.refresh()
            d = dataset[...].flatten()
            kds.append(d)

        return kds

    def _timeout(self):
        if not self.start_time:
            return False

        if time.time() > self.start_time + self.timeout:
            return True
        else:
            return False

    def _check_finished_dataset(self):
        if self.finished_dataset is None:
            return False

        f = self.h5file[self.finished_dataset]
        f.refresh()

        finished = f[0] == 1

        if self._prelim_finished_check and finished:
            return True

        # go through the timeout loop once more
        # in case finish is flushed slightly before
        # the last of the keys
        if finished:
            self._prelim_finished_check = True

        return False

    def is_finished(self):
        """Returns True if the KeyFollower instance has completed its iteration"""

        if self.current_key != self.current_max:
            return False

        if self._timeout():
            logger.debug("Finished on timeout")
            return True

        if self._check_finished_dataset():
            logger.debug("Finished on finished dataset check")
            return True

        return False

    def get_current_max(self):
        """Returns the current maximum key"""
        return self.current_max

    def refresh(self):
        """Force an update of the current maximum key"""
        return self._is_next()

    def are_keys_complete(self):
        """Check position of current maximum against the max shape"""
        return self.current_max == (np.prod(self.maxshape) - 1)


class RowKeyFollower:
    def __init__(
        self, h5file, keypaths, timeout=10, finished_dataset=None, row_size=None
    ):
        self.inner_key_follower = KeyFollower(
            h5file, keypaths, timeout=timeout, finished_dataset=finished_dataset
        )
        self.row_size = row_size
        self.scan_rank = -1
        self.maxshape = None
        self._row_count = -1

    def __iter__(self):
        return self

    def check_datasets(self):

        self.inner_key_follower.check_datasets()
        self.scan_rank = self.inner_key_follower.scan_rank
        self.maxshape = self.inner_key_follower.maxshape

        if self.row_size is None:
            rsize = self.inner_key_follower.maxshape[-1]
            if rsize is None:
                raise RuntimeError(
                    "Row size must be defined if fastest max shape dimension is -1"
                )

            self.row_size = rsize

    def __next__(self):

        for i in range(self.row_size - 1):
            next(self.inner_key_follower)

        return next(self.inner_key_follower)

    def reset(self):
        """Reset the iterator to start again from index 0"""
        self._row_count = -1
        self.inner_key_follower.reset()



from distutils.core import setup

# read the contents of your README file
from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name="swmr_tools",
    packages=["swmr_tools"],
    version="0.7.2",
    license="MIT",
    description="Python iterator for safely monitoring NeXus files",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Diamond Light Source Ltd",
    author_email="scientificsoftware@diamond.ac.uk",
    url="https://github.com/DiamondLightSource/python-swmrtools",
    keywords=["HDF5", "Iterator", "Diamond"],
    install_requires=["numpy", "h5py"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
)



from .keyfollower import KeyFollower
import logging
import numpy as np
from .utils import get_position, create_dataset, append_data
import sys

try:
    import blosc
except ImportError:
    pass

logger = logging.getLogger(__name__)


class DataSource:
    """Iterator for returning dataset frames for any number of datasets. This
    class is largely a wrapper around the KeyFollower and FrameReader classes.

    Parameters
    ----------

    h5file: h5py.File
        Instance of h5py.File object. Choose the file containing data you wish
        to follow.

    keypaths: list
        A list of paths (as strings) to key datasets in the hdf5 file. Can also be
        the path to groups, but the groups must contain only key datasets.

    dataset_paths: list
        A list of paths (as strings) to datasets in h5file that you wish
        to return frames from (Note: paths must be to the dataset and not
                               to the group containing it).

    timeout: int (optional)
        The maximum time allowed for a dataset to update before the timeout
        termination condition is triggered and iteration is halted. If a value
        is not set this will default to 10 seconds.

    finished_dataset: string (optional)
        Path to a scalar hdf5 dataset which is zero when the file is being
        written to and non-zero when the file is complete. Used to stop
        the iterator without waiting for the timeout.

    cache_dataset: bool (optional - deprecated)
        Deprecated - Now does nothing - will be removed in future release, not holding the dataset
        open leads to performance issues.

    use_direct_chunk: bool (optional)
        If dataset chunking is aligned to a single frame, and data is blosc
        compressed, will use direct chunk read and compression outside of h5py
        for performance.

    interleaved_paths: dict (optional)
        A dictionary of string to lists of dataset paths. Where frames are written by multiple file
        writers in an interleaved fashion, this will take frames alternating from each
        dataset in the list. Key dataset should be a VDS of interleaved keys. For use where
        direct chunk read is required but not possible through a VDS stack of frames. Only
        for stacks of frames.


    Examples
    --------

    >>> with h5py.File("/home/documents/work/data/example.h5", "r", swmr = True) as f:
    >>>     df = DataSource(f, path_to_key_group, path_to_datasets)
    >>>     for frame_dict in df:
    >>>         print(frame_dict)

    """

    def __init__(
        self,
        h5file,
        keypaths,
        dataset_paths,
        timeout=10,
        finished_dataset=None,
        cache_datasets=True,
        use_direct_chunk=False,
        interleaved_paths=None,
    ):
        self.h5file = h5file
        self.dataset_paths = dataset_paths
        self.interleaved_paths = interleaved_paths
        self.cache = {}
        self.interleaved_cache = {}
        self.kf = KeyFollower(h5file, keypaths, timeout, finished_dataset)
        self.kf.check_datasets()

        if dataset_paths is None and interleaved_paths is None:
            raise RuntimeError("No data specified to follow!")

        self._add_datasets_to_cache(use_direct_chunk)
        self._add_interleaved_datasets_to_cache(use_direct_chunk)

    def _add_datasets_to_cache(self, use_direct_chunk):
        if self.dataset_paths is not None:
            for path in self.dataset_paths:
                self.cache[path] = FrameReader(
                    path,
                    self.h5file,
                    self.kf.scan_rank,
                    use_direct_chunk=use_direct_chunk,
                )

    def _add_interleaved_datasets_to_cache(self, use_direct_chunk):
        if self.interleaved_paths is not None:
            for path, path_list in self.interleaved_paths.items():
                self.interleaved_cache[path] = []
                for ds_path in path_list:
                    fr = FrameReader(
                        ds_path,
                        self.h5file,
                        self.kf.scan_rank,
                        use_direct_chunk=use_direct_chunk,
                    )

                    self.interleaved_cache[path].append(fr)

    def __iter__(self):
        return self

    def __next__(self):

        if self.kf.is_finished():
            raise StopIteration

        else:

            current_dataset_index = next(self.kf)

            output = SliceDict()

            self._add_datasets_to_output(current_dataset_index, output)
            self._add_interleaved_datasets_to_output(current_dataset_index, output)

            return output

    def _add_datasets_to_output(self, current_dataset_index, output):
        if self.dataset_paths is None:
            return

        for path in self.dataset_paths:
            # If caching requested use from cache
            # else create
            fg = self.cache.get(path, FrameReader(path, self.h5file, self.kf.scan_rank))
            frame, slice_metadata = fg.read_frame(current_dataset_index)
            output[path] = frame
            if output.slice_metadata is None:
                output.slice_metadata = slice_metadata
                output.maxshape = self.kf.maxshape
                output.index = current_dataset_index

    def _add_interleaved_datasets_to_output(self, current_dataset_index, output):
        if self.interleaved_paths is None:
            return

        for path, frs in self.interleaved_cache.items():
            n_frs = len(frs)
            fr_index = current_dataset_index % (n_frs)

            frame, slice_metadata = frs[fr_index].read_frame(
                current_dataset_index // n_frs
            )
            output[path] = frame

            if output.slice_metadata is None:
                updated = (
                    slice(current_dataset_index, current_dataset_index + 1),
                    *slice_metadata[1:],
                )
                output.slice_metadata = updated
                output.maxshape = self.kf.maxshape
                output.index = current_dataset_index

    def reset(self):
        """Reset the iterator to start again from frame 0"""
        self.kf.reset()

    def create_dataset(self, data, fh, path):
        scan_max = self.kf.maxshape
        return create_dataset(data, scan_max, fh, path)

    def append_data(self, data, slice_metadata, dataset):
        return append_data(data, slice_metadata, dataset)

    def is_scan_finished(self):
        return self.kf._prelim_finished_check


class SliceDict(dict):
    """Dictionary with attributes for the slice metadata and maxshape of the scan"""

    def __init__(self, *args, **kw):
        super(SliceDict, self).__init__(*args, **kw)
        self.slice_metadata = None
        self.maxshape = None
        self.index = None


class FrameReader:
    """Class for extracting frames from a dataset given an index generated by
    from an instance of the KeyFollower class

    Parameters
    ----------
    h5file : h5py.File
        Instance of h5py.File object. Choose the file containing dataset you
        want to extract frames from.

    dataset : str
        The full path to the dataset to extract frames from in the hdf5_File

    scan_rank: int
        The rank of the "non-data-frame" part of the N-dimensional dataset
        for example 1 if the dataset is rank 3 and the frames are images (2D) or
        2 if the dataset is rank 3 and the frames are spectra/patterns/vectors
        (i.e. a grid scan). The KeyFollower class has the scan_rank as an attribute
        that can be used here.

    Examples
    --------

    >>> #open and hdf5 file using context manager
    >>> with h5py.File("/path/to/file") as f:
    >>>     fr = FrameReader(f, "/path/to/dataset", 2)
    >>>     #call methods on fg to get the data that you want

    """

    def __init__(self, dataset, h5file, scan_rank, use_direct_chunk=False):
        self.dataset = dataset
        self.h5file = h5file
        self.scan_rank = scan_rank
        self.use_direct_chunk = use_direct_chunk
        self.ds = self.h5file[self.dataset]

        if use_direct_chunk:
            self.use_direct_chunk = False
            prop_dcid = self.ds.id.get_create_plist()
            if prop_dcid.get_nfilters() == 1 and prop_dcid.get_filter(0)[0] == 32001:
                chunk = prop_dcid.get_chunk()
                shape = self.ds.shape
                frame_rank = len(shape) - scan_rank
                if shape[-frame_rank:] == chunk[-frame_rank:] and all(
                    [i == 1 for i in chunk[:scan_rank]]
                ):
                    if "blosc" in sys.modules:
                        self.use_direct_chunk = True
                        self.chunk = chunk

    def read_frame(self, index):
        """Method for using an index from KeyFollower to extract that frame
        from the chosen hdf5 dataset.

        Parameters
        ----------
        index : int
            Index for the correspondng non-zero unique key generated by an
            instance of KeyFollower.Follower


        Examples
        --------

        >>> #create a list of frames from all the keys returned by an instance
        >>> #of KeyFollower.Follower
        >>> frame_list = []
        >>> with h5py.File("path/to/file") as f:
        >>>     fg = KeyFollower.FrameGrabber(f, "path/to/dataset")
        >>>     kf = KeyFollower.Follower(f, ["path/to/keys_1",
        >>>                          "path/to/keys_2"],
        >>>                      timeout = 10)
        >>>     for key in kf:
        >>>         frame = fg.Grabber(key)
        >>>         frame_list.append(frame)
        """

        ds = self.ds if self.ds is not None else self.h5file[self.dataset]
        shape = ds.shape

        try:
            # might fail if dataset is cached
            pos = self.get_pos(index, shape)
        except ValueError:
            # refresh dataset and try again
            if hasattr(ds, "refresh"):
                ds.refresh()

            shape = ds.shape
            pos = self.get_pos(index, shape)

        rank = len(shape)

        slices = [slice(0, None, 1)] * rank

        for i in range(len(pos)):
            slices[i] = slice(pos[i], pos[i] + 1)

        if self.use_direct_chunk:
            return self.get_frame_direct(ds, pos, rank, slices)
        else:
            return self.get_frame(ds, pos, rank, slices)

    def get_frame(self, ds, pos, rank, slices):
        frame = ds[tuple(slices)]
        return frame, tuple(slices[: self.scan_rank])

    def get_frame_direct(self, ds, pos, rank, slices):
        chunk_pos = [0] * rank
        for i in range(len(pos)):
            chunk_pos[i] = pos[i]

        out = ds.id.read_direct_chunk(chunk_pos)
        decom = blosc.decompress(out[1])
        a = np.frombuffer(decom, dtype=ds.dtype, count=-1)
        return a.reshape(self.chunk), tuple(slices[: self.scan_rank])

    def get_pos(self, index, shape):
        return get_position(index, shape, self.scan_rank)
